"""
The RPC(Remote Procedure Call) module.

Contain RPC utils, for easy of use.

"""
import datetime
import json
import os
import random
import string

from jsonrpcclient import requests, response

from . import message_data

JSONRPC = '2.0'

session_id = "".join(random.choices(string.ascii_letters + string.digits, k=10))
session_count = 0


def id_count(verb):
    """Keep track of the trace-id."""
    global session_count
    session_count += 1
    return f"{session_id}_{verb}_{session_count}"


def create_meta(network, network_id):
    """
    Create meta.

    Creates and returns a meta data about a network.

    Args:
        network:
        network_id: Identify number of a network.

    Returns:
        meta data dictionary.

    """
    return {
        'id': network_id,
        'type': '{}'.format(network),
        'version': '2.0'
    }


def get_rpc_success_response(message_id):
    """
    Get success response.

    Returns a success, JSON formatted, encoded in utf-8 response.

    Args:
        message_id: Id of a message.

    Returns:
        JSON formatted data.

    """
    success_response = str(response.SuccessResponse(jsonrpc=JSONRPC,
                                                    id=message_id,
                                                    result=True))
    return json.loads(success_response)


def get_rpc_fail_response(message_id, text):
    """
    Get fail response.

    Returns a fail, JSON formatted, encoded in utf-8 response with a error
    message.

    Args:
        message_id: Id of a message.
        text: A message of error description.

    Returns:
        JSON formatted data.

    """
    error_description = {'message': text, 'code': -32020, 'data': ''}
    error_response = str(response.ErrorResponse(jsonrpc=JSONRPC,
                                                id=message_id,
                                                error=error_description))
    return json.loads(error_response)


def time_stamp():
    """Return The default timestamp used for Wappsto."""
    return datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S.%fZ')


def is_upgradable():
    """
    Check if System is set to upgradable.

    Returns:
        True, if the system is upgradable,
        False, if it is not.

    """
    return os.getenv("UPGRADABLE", "").upper() == "TRUE"


def get_rpc_network(network_id, network_name, verb,
                    trace_id=None):
    """
    Retrieve network from server.

    Retrieves network from a server. The method is used while adding a
    network. Depending on request method it fills data_json_rpc dictionary
    with different information.

    Args:
        network_id: Unique identifying number of a network
        network_name: Name of a network
        verb: indicates what verb should be used.
        trace_id:  ID of the debug trace {default: None}

    Returns:
        JSON formatted data of network

    """
    meta = create_meta('network', network_id)
    if is_upgradable():
        meta.update({'upgradable': True})
    data_inside = {
        'meta': meta,
        'name': network_name
    }
    url = '/network'

    if verb == message_data.PUT:
        url = '/{}'.format(network_id)

    if trace_id:
        url = "{}?trace={}".format(url, trace_id)

    data_json_rpc = requests.Request(
        verb,
        url=url,
        data=data_inside,
        request_id=id_count(verb)
    )

    return data_json_rpc


def get_rpc_state(
        data,
        network_id,
        device_id,
        value_id,
        state_id,
        set_type,
        verb,
        trace_id=None
):
    """
    Retrieve state of the value.

    Retrieves the value state from a server. Uses functionality of
    base_device_value to get a directory with common attributes of value
    class.

    Args:
        data: Data from the state of the value.
        network_id: Unique identifier of the network.
        device_id: Unique identifier of a device.
        value_id: Unique identifier of a value.
        state_id: Unique identifier of the state.
        set_type: The type to set.
        verb: indicates what verb should be used.
        trace_id: ID of the debug trace. (default: {None})

    Returns:
        JSON formatted data of the state.

    """
    if verb != message_data.GET:
        device_state = {
            'meta': create_meta('state', state_id),
            'type': set_type,
            'status': 'Send',
            'data': data,
            'timestamp': time_stamp()
        }
    else:
        device_state = None

    url = '/network/{}/device/{}/value/{}/state/{}'
    url = url.format(network_id, device_id, value_id, state_id)

    if trace_id:
        url = '{}?trace={}'.format(url, trace_id)

    data_json_rpc = requests.Request(
        verb,
        url=url,
        data=device_state,
        request_id=id_count(verb)
    )
    return data_json_rpc


def get_rpc_delete(network_id,
                   device_id,
                   value_id,
                   state_id,
                   trace_id=None):
    """
    Creates delete request.

    The method is used to create message that could allow to
    delete network or its elements.

    Args:
        network_id: id of the network to delete/modify.
        device_id: id of the device to delete/modify.
        value_id: id of the value to delete/modify.
        state_id: id of the state to delete.
        trace_id: ID of the debug trace. {default: None}

    Returns:
        JSON formatted data of delete message

    """
    # UNSURE(MBK): I feel that there is a more pythonic way of doing this.
    if network_id:
        url = '/network/{}'.format(network_id)
        if device_id:
            url += '/device/{}'.format(device_id)
            if value_id:
                url += '/value/{}'.format(value_id)
                if state_id:
                    url += '/state/{}'.format(state_id)

    if trace_id:
        url = '{}?trace={}'.format(url, trace_id)

    data_json_rpc = requests.Request(
        'DELETE',
        url=url,
        request_id=id_count('DELETE')
    )

    return data_json_rpc


def get_rpc_whole_json(json_data, trace_id=None):
    """
    Creates request containing the whole json file.

    The method is used while starting initializing and it fills
    data_json_rpc dictionary with url and data containing json object.

    Args:
        json_data: Data read from json file.
        trace_id:  ID of the debug trace {default: None}

    Returns:
        JSON formatted data of network

    """
    url = '/{}'.format("network")

    if trace_id:
        url = "{}?trace={}".format(url, trace_id)

    data_json_rpc = requests.Request(
        'POST',
        url=url,
        data=json_data,
        request_id=id_count('POST')
    )
    return data_json_rpc



"""
The wappsto decoding module.

Handles decoding object instances from a JSON file.
"""
import logging
from ..modules import network as network_module
from ..modules import device as device_module
from ..modules import value as value_module
from ..modules import state as state_module


class WappstoDecoder:
    """
    The wappsto decoding class.

    Handles decoding JSON information into object instances. This
    allows the system to use the provided information.
    """

    def __init__(self):
        """
        Initialize WappstoDecoder.

        Initializes the WappstoDecoder class.
        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

    def decode_network(self, json_data, data_manager):
        """
        Decode instance of Network class.

        Handles the decoding of the network instance, contains a template to
        decode the network with.

        Args:
            json_data: Dictionary object from what data must be extracted.
            data_manager: Reference to the instance of the DataManager class.

        Returns:
            Network object.

        """
        network = network_module.Network(
            uuid=json_data.get('meta').get('id'),
            version=json_data.get('meta').get('version'),  # TODO (aeb): Remove? Does not make much sense
            name=json_data.get('name'),
            devices=[],
            data_manager=data_manager
        )
        network.devices = self.decode_device(json_data, network)

        self.wapp_log.debug("Network {} built.".format(network))
        return network

    def decode_device(self, json_data, parent):
        """
        Decode instance of Device class.

        Handles the decoding of the device instance, contains a template to
        decode the device with.

        Args:
            json_data: Dictionary object used for data extraction.
            parent: Reference to the instance of the Network class.

        Returns:
            List of devices.

        """
        devices = []
        for device_iterator in json_data.get('device', []):
            device = device_module.Device(
                parent=parent,
                uuid=device_iterator.get('meta').get('id'),
                name=device_iterator.get('name'),
                product=device_iterator.get('product'),
                protocol=device_iterator.get('protocol'),
                serial_number=device_iterator.get('serial'),
                version=device_iterator.get('version', ""),
                manufacturer=device_iterator.get('manufacturer'),
                communication=device_iterator.get('communication'),
                description=device_iterator.get('description')
            )
            device.values = self.decode_value(device_iterator, device)
            devices.append(device)

            msg = "Device {} appended to {}".format(device, devices)
            self.wapp_log.debug(msg)
        return devices

    def decode_value(self, json_data, parent):
        """
        Decode instance of Value class.

        Handles the decoding of the value instance, contains a template to
        decode the value with.

        Args:
            json_data: Dictionary object used for data extraction.
            parent: Reference to the instance of the Device class.

        Returns:
            List of values.

        """
        values = []
        for value_iterator in json_data.get('value', []):
            data_type = None
            number_min = None
            number_max = None
            number_step = None
            number_unit = None
            string_encoding = None
            string_max = None
            blob_encoding = None
            blob_max = None

            if 'string' in value_iterator:
                data_type = 'string'
                string_encoding = value_iterator.get('string').get('encoding')
                string_max = value_iterator.get('string').get('max')
            elif 'blob' in value_iterator:
                data_type = 'blob'
                blob_encoding = value_iterator.get('blob').get('encoding')
                blob_max = value_iterator.get('blob').get('max')
            elif 'number' in value_iterator:
                data_type = 'number'
                number_min = value_iterator.get('number').get('min')
                number_max = value_iterator.get('number').get('max')
                number_step = value_iterator.get('number').get('step')
                number_unit = value_iterator.get('number').get('unit')

            value = value_module.Value(
                parent=parent,
                uuid=value_iterator.get('meta').get('id'),
                name=value_iterator.get('name'),
                type_of_value=value_iterator.get('type'),
                data_type=data_type,
                permission=value_iterator.get('permission'),
                number_max=number_max,
                number_min=number_min,
                number_step=number_step,
                number_unit=number_unit,
                string_encoding=string_encoding,
                string_max=string_max,
                blob_encoding=blob_encoding,
                blob_max=blob_max,
                period=value_iterator.get('period', None),
                delta=value_iterator.get('delta', None)
            )
            for state in self.decode_state(value_iterator, value):
                if state.state_type == 'Report':
                    value.add_report_state(state)
                elif state.state_type == 'Control':
                    value.add_control_state(state)
            values.append(value)

            msg = "Value {} appended to {}".format(value, values)
            self.wapp_log.debug(msg)
        return values

    def decode_state(self, json_data, parent):
        """
        Decode instance of State class.

        Handles the decoding of the state instance, contains a template to
        decode the state with.

        Args:
            json_data: Dictionary object used for data extraction.
            parent: Reference to the instance of the Value class.

        Returns:
            List of states.

        """
        states = []
        for state_iterator in json_data.get('state', []):
            state = state_module.State(
                parent=parent,
                uuid=state_iterator.get('meta').get('id'),
                state_type=state_iterator.get('type'),
                timestamp=state_iterator.get('timestamp'),
                init_value=state_iterator.get('data')
            )
            states.append(state)
        return states



"""
The json data managing module.

Saves and loads data from json files.
"""
import os
import json
import logging
import warnings
from . import encoder
from . import decoder


class DataManager:
    """
    The data manager class.

    Reads data from Json file and saves current state of the whole network.
    """

    def __init__(
            self,
            json_file_name,
            load_from_state_file,
            path_to_calling_file):
        """
        Initialize the DataManager class.

        Initializes the DataManager class which reads data from Json file and
        saves the state of the whole network.

        Args:
            json_file_name: The name of the JSON file to parse.
            load_from_state_file: A True/False flag to denote whether to load
                from the saved files directory.
            path_to_calling_file: The path to files location.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.wappsto_encoder = encoder.WappstoEncoder()
        self.wappsto_decoder = decoder.WappstoDecoder()

        self.path_to_calling_file = path_to_calling_file
        self.json_file_name = os.path.join(self.path_to_calling_file, json_file_name)

        if load_from_state_file:
            json_file_name = self.get_latest_instance()
            if json_file_name is not None:
                self.json_file_name = json_file_name
        self.read_file()

    def __getattr__(self, attr):
        """
        Get attribute value.

        When trying to get value from outdated attribute warning is raised about
        it being deprecated and the new variable is used instead.

        Returns:
            data stored in the new location

        """
        if attr in ["device_list"]:
            warnings.warn("Property {} is deprecated".format(attr))
            return self.network.devices
        if attr in ["network_cl"]:
            warnings.warn("Property {} is deprecated".format(attr))
            return self.network

    def get_latest_instance(self):
        """
        Gets latest saved instance.

        Gets all files from instance folder and returns the most recently changed.

        Returns:
            name of the most recently changed file

        """
        path = os.path.join(self.path_to_calling_file, 'saved_instances')
        os.makedirs(path, exist_ok=True)

        file_paths = []
        for file_name in os.listdir(path):
            file_paths.append(os.path.join(path, file_name))

        latest_file = None
        if len(file_paths) > 0:
            latest_file = str(max(file_paths, key=os.path.getctime))
            self.wapp_log.debug('Latest file: {}'.format(latest_file))

        return latest_file

    def read_file(self):
        """
        Reads file.

        Reads file with the provided name and calls method which parses it.

        """
        try:
            with open(self.json_file_name) as data_file:
                file_data = data_file.read()
                self.parse_json_file(file_data)
            self.wapp_log.debug("Opening file: {}".format(self.json_file_name))
        except FileNotFoundError as fnfe:
            self.wapp_log.error("Error finding file: {}".format(fnfe))
            raise fnfe

    def parse_json_file(self, file_data):
        """
        Parses json file.

        Retrieves the necessary data from the file and calls decode_network,
        results of which are saved in network variable.

        Args:
            file_data: string data read from the file.

        """
        try:
            self.decoded = json.loads(file_data)
            json_container = self.decoded.get('data')
            if not isinstance(json_container, dict):
                json_container = json.loads(json_container)
        except json.JSONDecodeError as jde:
            self.wapp_log.error("Error decoding: {}".format(jde))
            raise jde

        self.wapp_log.debug("RAW JSON DATA:\n\n{}\n\n".format(json_container))

        self.network = self.wappsto_decoder.decode_network(json_container, self)

    def save_instance(self):
        """
        Saves current instance of the whole network.

        Encodes the whole network and saves it in the saved instance folder.

        """
        encoded_string = str(self.get_encoded_network())
        encoded_string = encoded_string.replace("\'", "\\\"")
        encoded_string = '{{"data":"{}"}}'.format(encoded_string)

        path = os.path.join(self.path_to_calling_file, 'saved_instances')
        os.makedirs(path, exist_ok=True)
        json_base_name = os.path.basename(self.json_file_name)
        path_open = os.path.join(path, json_base_name)

        with open(path_open, "w+") as network_file:
            network_file.write(encoded_string)

        msg = "Saved {} to {}".format(encoded_string, network_file)
        self.wapp_log.debug(msg)

    def get_by_id(self, id):
        """
        Wappsto get by id.

        Retrieves the instance of a class if its id matches the provided one.

        Args:
            id: unique identifier used for searching

        Returns:
            A reference to the network/device/value/state object instance.

        """
        # UNSURE(MBK): I feel that there is a more pythonic way of doing this.
        message = "Found instance of {} object with id: {}"
        if self.network is not None:
            if self.network.uuid == id:
                self.wapp_log.debug(message.format("network", id))
                return self.network

            for device in self.network.devices:
                if device.uuid == id:
                    self.wapp_log.debug(message.format("device", id))
                    return device

                for value in device.values:
                    if value.uuid == id:
                        self.wapp_log.debug(message.format("value", id))
                        return value

                    if value.control_state is not None and value.control_state.uuid == id:
                        self.wapp_log.debug(message.format("control state", id))
                        return value.control_state

                    if value.report_state is not None and value.report_state.uuid == id:
                        self.wapp_log.debug(message.format("report state", id))
                        return value.report_state

        self.wapp_log.warning("Failed to find object with id: {}".format(id))

    def get_encoded_network(self):
        """
        Gets encoded network.

        Gets dictionary representing current state of the whole network.

        """
        return self.wappsto_encoder.encode_network(self.network)



"""
The send_data module.

Handles sending data to the server.

"""
import json
import logging
import random
import ssl
import threading

import urllib.request as request

from . import message_data
from . import seluxit_rpc


MAX_BULK_SIZE = 10
t_url = 'https://tracer.iot.seluxit.com/trace?id={}&parent={}&name={}&status={}'  # noqa: E501


class SendData:
    """The SendData class that handles sending information."""

    def __init__(self, client_socket, automatic_trace):
        """
        Initialize the SendData class.

        Initializes an object of SendData class by passing required
        parameters. While initialization, wapp_log is created.

        Args:
            client_socket: reference to ClientSocket instance.
            automatic_trace: indicates if all messages automaticaly send trace.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        self.client_socket = client_socket
        self.automatic_trace = automatic_trace
        self.add_trace_to_report_list = {}
        self.bulk_send_list = []
        self.lock = threading.Lock()

    def create_trace(self, parent, trace_id=None):
        """
        Creates trace.

        Creates trace if necessary, by using generated data and existing
        information.

        Args:
            parent: owner of trace.
            trace_id: existing id used for tracing.

        Returns:
            trace id.

        """
        if self.automatic_trace and trace_id is None:
            random_int = random.randint(1, 25000)
            control_value_id = "{}{}".format(self.client_socket.data_manager.network.name,
                                             random_int)

            trace_id = random_int

            trace = message_data.MessageData(
                message_data.SEND_TRACE,
                parent=parent,
                trace_id=trace_id,
                data=None,
                text="ok",
                control_value_id=control_value_id)
            self.send_trace(trace)
        return trace_id

    def create_bulk(self, data):
        """
        Creates bulk message.

        Accomulates all messages in one and once sending_queue is empty or
        bulk limit is reached it is sent.

        Args:
            data: JSON communication message data.

        """
        self.lock.acquire()
        if data is not None:
            self.bulk_send_list.append(data)

        # Logic Checks
        empty_q = self.client_socket.sending_queue.qsize() == 0
        bulk_maxed = len(self.bulk_send_list) >= MAX_BULK_SIZE

        if (empty_q or bulk_maxed):
            free_con_lines = 10 - len(self.client_socket.packet_awaiting_confirm)
            send_size = len(self.bulk_send_list) if len(self.bulk_send_list) <= free_con_lines else free_con_lines
            self.wapp_log.debug("Starting to sending bulk data.")
            self.send_data([self.bulk_send_list.pop(0) for _ in range(send_size)])
        self.lock.release()

    def send_data(self, data):
        """
        Send JSON data.

        Sends the encoded JSON message through the socket.

        Args:
            data: JSON communication message data.

        """
        try:
            for data_element in data:
                self.client_socket.get_object_without_none_values(data_element)
                if len(data_element) == 0:
                    self.wapp_log.debug('Removing data element of length 0: {}'.format(data_element))
                    data.remove(data_element)  # WARNING(MBK): Do not remove from list when looping over it.

            if self.client_socket.connected:
                for data_element in data:
                    if data_element.get("method", "") in ["PUT", "POST", "DELETE"]:
                        self.client_socket.add_id_to_confirm_list(data_element)
                if len(data) > 0:
                    data = json.dumps(data)
                    data = data.encode('utf-8')
                    self.client_socket.my_socket.sendall(data)
                    self.wapp_log.debug('Raw Json Sent: {}'.format(data))
            else:
                self.wapp_log.warning("Data added to storage!")
                self.client_socket.event_storage.add_message(data)
        except OSError as e:
            self.wapp_log.warning("Data added to storage!")
            self.client_socket.event_storage.add_message(data)

            self.client_socket.connected = False
            msg = "Error sending: {}".format(e)
            self.wapp_log.error(msg)
            self.client_socket.request_reconnect()

    def send_thread(self):
        """
        Create a send thread.

        Retrieves packages from the sending queue to
        send data.
        """
        self.wapp_log.debug("SendingThread Started!")

        while True:
            package = self.client_socket.sending_queue.get()
            if package.msg_id == message_data.SEND_SUCCESS:
                self.send_success(package)

            elif package.msg_id == message_data.SEND_FAILED:
                self.send_failed(package)

            elif package.msg_id == message_data.SEND_REPORT:
                self.send_report(package)

            elif package.msg_id == message_data.SEND_CONTROL:
                self.send_control(package)

            elif package.msg_id == message_data.SEND_DELETE:
                self.send_delete(package)

            elif package.msg_id == message_data.SEND_TRACE:
                self.send_trace(package)

            elif package.msg_id == message_data.SEND_RECONNECT:
                self.send_reconnect(package)

            elif package.msg_id == message_data.POKE:
                self.wapp_log.debug("Was Poked.")
                self.create_bulk(None)

            else:
                self.wapp_log.warning("Unhandled send")

            self.client_socket.sending_queue.task_done()

    def send_success(self, package):
        """
        Send a success message.

        Sends a message to notify of a successful.

        Args:
            package: A sending queue item.

        """
        self.wapp_log.info("Sending success for ID: {}".format(package.rpc_id))
        rpc_success_response = seluxit_rpc.get_rpc_success_response(
            package.rpc_id
        )
        self.create_bulk(rpc_success_response)

    def send_failed(self, package):
        """
        Send a fail message.

        Sends a message to notify about a sending failure.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending failed for ID: {}".format(package.rpc_id))
        self.wapp_log.info("Sending failed reason: {}".format(package.text))
        rpc_fail_response = seluxit_rpc.get_rpc_fail_response(
            package.rpc_id,
            package.text
        )
        self.create_bulk(rpc_fail_response)

    def send_report(self, package):
        """
        Send a report.

        Sends a report message from the package.

        Args:
            package: A sending queue item.

        """
        self.wapp_log.info("Sending report message")
        if not package.trace_id:
            if package.value_id in self.add_trace_to_report_list.keys():
                package.trace_id = (
                    self.add_trace_to_report_list.pop(package.value_id)
                )

        package.trace_id = self.create_trace(
            package.network_id, package.trace_id)

        local_data = seluxit_rpc.get_rpc_state(
            package.data,
            package.network_id,
            package.device_id,
            package.value_id,
            package.state_id,
            'Report',
            package.verb,
            trace_id=package.trace_id
        )
        self.create_bulk(local_data)
        data_decoded = local_data.get('params').get('data').get('data')
        self.wapp_log.info('Report value send: {}'.format(data_decoded))

    def send_control(self, package):
        """
        Send data handler.

        Sends the data from outgoing control messages to the appropriate
        handler method.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending control message")
        package.trace_id = self.create_trace(
            package.network_id, package.trace_id)

        local_data = seluxit_rpc.get_rpc_state(
            package.data,
            package.network_id,
            package.device_id,
            package.value_id,
            package.state_id,
            'Control',
            package.verb,
            trace_id=package.trace_id
        )
        self.create_bulk(local_data)

    def send_delete(self, package):
        """
        Send data delete request.

        Sends the data to be deleted.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending delete message")
        package.trace_id = self.create_trace(
            package.network_id, package.trace_id)

        local_data = seluxit_rpc.get_rpc_delete(
            package.network_id,
            package.device_id,
            package.value_id,
            package.state_id,
            package.trace_id
        )
        self.create_bulk(local_data)

    def send_trace(self, package):
        """
        Send data trace.

        Provides a trace URL for debugging purposes.

        Args:
            package: Sending queue item.

        """
        if package.control_value_id:
            control_value_id = package.control_value_id
            self.add_trace_to_report_list[control_value_id] = package.trace_id

        attempt = str(t_url).format(
            package.trace_id,
            package.parent,
            package.data,
            package.text
        )

        context = ssl._create_unverified_context()
        trace_req = request.urlopen(attempt, context=context)
        msg = "Sending tracer https message {} response {}"
        msg = msg.format(attempt, trace_req.getcode())
        self.wapp_log.debug(msg)

    def send_reconnect(self, package):
        """
        Send a reconnect attempt.

        Sends a request to attempt to reconnect to the server.

        Args:
            package: Sending queue item.

        """
        self.wapp_log.info("Sending reconnect data")
        package.trace_id = self.create_trace(
            package.network_id, package.trace_id)

        rpc_network = seluxit_rpc.get_rpc_network(
            self.client_socket.data_manager.network.uuid,
            self.client_socket.data_manager.network.name,
            package.verb,
            trace_id=package.trace_id
        )
        # NOTE: Need to be send even if bulk is full.
        self.send_data([rpc_network])
        self.wapp_log.info("Reconnect data send")
        self.wapp_log.info("Sending awaiting data")
        self.send_data(list(self.client_socket.packet_awaiting_confirm.values()))
        self.wapp_log.info("Awaiting data send")



"""
The value module.

Stores attributes for the value instance and handles value-related
methods.
"""
import logging
import threading
import warnings

from ..connection import message_data
from ..connection import seluxit_rpc
from ..errors import wappsto_errors


def isNaN(num):
    """Test if input is a float 'NaN' value."""
    return num != num


class Value:
    """
    Value instance.

    Stores attributes for the value instance and handles value-related
    methods.
    """

    def __init__(
        self,
        parent,
        uuid,
        name,
        type_of_value,
        data_type,
        permission,
        number_max,
        number_min,
        number_step,
        number_unit,
        string_encoding,
        string_max,
        blob_encoding,
        blob_max,
        period,
        delta
    ):
        """
        Initialize the Value class.

        Initializes an object of value class by passing required parameters.

        Args:
            parent: Reference to a device object
            uuid: An unique identifier of a device
            name: A name of a device
            type_of_value: Determines a type of value [e.g temperature, CO2]
            data_type: Defines whether a value is string, blob or number
            permission: Defines permission [read, write, read and write]
            (if data_type is number then these parameters are relevant):
            number_max: Maximum number a value can have
            number_min: Minimum number a value can have
            number_step: Number defining a step
            number_unit: Unit in which a value should be read
            (if data_type is string then these parameters are irrelevant):
            string_encoding: A string encoding of a value
            string_max: Maximum length of string
            (if data_type is blob then these parameters are irrelevant):
            blob_encoding: A blob encoding of a value
            blob_max: Maximum length of a blob

            period: defines the time after which a value should send report
                message. Default: {None})
            delta: defines the a difference of value (default: {None})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.parent = parent
        self.uuid = uuid
        self.name = name
        self.type_of_value = type_of_value
        self.data_type = data_type
        self.permission = permission
        # The value shared between state instances.
        self.number_max = number_max
        self.number_min = number_min
        self.number_step = number_step
        self.number_unit = number_unit
        self.string_encoding = string_encoding
        self.string_max = string_max
        self.blob_encoding = blob_encoding
        self.blob_max = blob_max
        self.report_state = None
        self.control_state = None
        self.callback = None

        self.timer = threading.Timer(None, None)
        self.last_update_of_report = None

        # if self._invalid_step(self.number_max):
        #     msg = "Inconsistent max, min & step provided. "
        #     msg += "'(max-min)/step' do not appear to an integer-like."
        #     self.wapp_log.warning(msg)

        if period:
            self.set_period(period)
        if delta:
            self.set_delta(delta)

        msg = "Value {} debug: {}".format(name, str(self.__dict__))
        self.wapp_log.debug(msg)

    def __getattr__(self, attr):  # pragma: no cover
        """
        Get attribute value.

        When trying to get value from last_controlled warning is raised about
        it being deprecated and calls get_data instead.

        Returns:
            value of get_data

        """
        if attr in ["last_controlled"]:
            warnings.warn("Property {} is deprecated".format(attr))
            return self.get_control_state().data

    def set_period(self, period):
        """
        Set the value reporting period.

        Sets the time defined in second to report a value to
        the server and starts timer.

        Args:
            period: Reporting period.

        """
        if period is None:
            self.wapp_log.warning("Period value is not provided.")
            return

        try:
            period = int(period)
        except ValueError:
            self.wapp_log.error("Period value must be a number.")
            return

        if period < 0:
            self.wapp_log.warning("Period value must not be lower then 0.")
            return

        self.period = period

    def enable_period(self):
        """
        Enable the Period handling if period was set.

        Enable the Period starts the timer that ensures that the
        value are getting updated with the right Periods.
        """
        if self.period is None:
            self.wapp_log.debug("Period was not set.")
            return
        if self.get_report_state() is not None:
            self.__set_timer()
            self.wapp_log.debug("Period successfully set.")
        else:
            self.wapp_log.warning("Cannot set the period for this value.")

    def __set_timer(self):
        """
        Set timer.

        Stop previous timer and sets new one if period value is not None.

        """
        self.timer.cancel()
        if self.period is not None:
            self.timer_elapsed = False
            self.timer = threading.Timer(self.period, self.__timer_done)
            self.timer.start()

    def __timer_done(self):
        self.__set_timer()
        self.timer_elapsed = True
        # self.handle_refresh()  # ERROR: Trickered double sampling. Text needed.

    def set_delta(self, delta):
        """
        Set the delta to report between.

        Sets the delta (range) of change to report in. When a change happens
        in the range of this delta it will be reported.

        Args:
            delta: Range to report between.

        """
        if delta is None:
            self.wapp_log.warning("Delta value is not provided.")
            return

        try:
            delta = float(delta)
        except ValueError:
            self.wapp_log.error("Delta value must be a number")
            return

        if delta < 0:
            self.wapp_log.warning("Delta value must not be lower then 0.")
            return

        if self.__is_number_type():
            self.delta = delta

    def enable_delta(self):
        """
        Enable the Delta handling, if delta is set.

        Enable the Delta, ATM do not do anything, other the inform
        if delta will be able to work.
        """
        if self.delta is None:
            self.wapp_log.debug("Delta was not set.")
            return
        if self.get_report_state():
            self.wapp_log.debug("Delta successfully set.")
        else:
            self.wapp_log.warning("Cannot set the delta for this value.")

    def get_parent_device(self):  # pragma: no cover
        """
        Retrieve parent device reference.

        Gets a reference to the device that owns this device.

        Returns:
            Reference to instance of Device class that owns this Value.

        """
        return self.parent

    def add_report_state(self, state):
        """
        Set report state reference to the value list.

        Adds a report state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.report_state = state
        msg = "Report state {} has been added.".format(state.parent.name)
        self.enable_period()
        self.enable_delta()
        self.wapp_log.debug(msg)

    def add_control_state(self, state):
        """
        Set control state reference to the value list.

        Adds a control state reference to the Value class.

        Args:
            state: Reference to instance of State class.

        """
        self.control_state = state
        msg = "Control state {} has been added".format(state.parent.name)
        self.wapp_log.debug(msg)

    def get_report_state(self):
        """
        Retrieve child report state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.report_state is not None:
            return self.report_state

        msg = "Value {} has no report state.".format(self.name)
        self.wapp_log.warning(msg)

    def get_control_state(self):
        """
        Retrieve child control state reference.

        Gets a reference to the child State class.

        Returns:
            Reference to instance of State class.

        """
        if self.control_state is not None:
            return self.control_state

        msg = "Value {} has no control state.".format(self.name)
        self.wapp_log.warning(msg)

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def _validate_value_data(self, data_value, err_msg=None):
        # TODO(MBK): Need refactoring, so it also nicely can be used for
        #            control validation, in 'receive_Data/incoming_put'
        if err_msg is None:
            err_msg = []
        if self.__is_number_type():
            try:
                if self._outside_range(data_value):
                    msg = "Invalid number. Range: {}-{}. Yours is: {}".format(
                        self.number_min,
                        self.number_max,
                        data_value
                    )
                    err_msg.append(msg)
                    self.wapp_log.warning(msg)
                if self._invalid_step(data_value):
                    msg = "Invalid Step. Step: {}. Min: {}. Value: {}".format(
                        self.number_step,
                        self.number_min,
                        data_value
                    )
                    err_msg.append(msg)
                    self.wapp_log.warning(msg)
                return str(data_value)
            except ValueError:
                msg = "Invalid type of value. Must be a number: {}"
                msg = msg.format(data_value)
                err_msg.append(msg)
                self.wapp_log.error(msg)
                return "NA"

        elif self.__is_string_type():
            if self.string_max is None:
                return data_value

            if len(str(data_value)) <= int(self.string_max):
                return data_value

            msg = "Value for '{}' not in correct range: {}."
            msg = msg.format(self.name, self.string_max)
            err_msg.append(msg)
            self.wapp_log.warning(msg)

        elif self.__is_blob_type():
            if self.blob_max is None:
                return data_value

            if len(str(data_value)) <= int(self.blob_max):
                return data_value

            msg = "Value for '{}' not in correct range: {}."
            msg = msg.format(self.name, self.blob_max)
            err_msg.append(msg)
            self.wapp_log.warning(msg)

        else:
            msg = "Value type '{}' is invalid".format(self.date_type)
            err_msg.append(msg)
            self.wapp_log.error(msg)

    def _outside_range(self, value):
        """
        Check weather or not the value are outside range.

        Args:
            value: The value to be checked.

        Returns:
            True, if outside range.
            False if inside range.
        """
        return not (self.number_min <= float(value) <= self.number_max)

    def _invalid_step(self, value):
        """
        Check weather or not the value are invalid step size.

        Args:
            value: The value to be checked.

        Returns:
            True, if invalid step size.
            False if valid step size.
        """
        x = (float(value) - self.number_min) / self.number_step
        return not (abs(round(x) - x) <= 1e-9)

    def update(self, data_value, timestamp=None):
        """
        Update value.

        Check if value has a state and validates the information in data_value
        if both of these checks pass then method send_state is called.

        Args:
            data_value: the new value.
            timestamp: time of action.

        Returns:
            True/False indicating the result of operation.

        """
        self._update_delta_period_values(data_value)

        if timestamp is None:
            timestamp = seluxit_rpc.time_stamp()

        state = self.get_report_state()
        if state is None:
            self.wapp_log.warning("Value is write only.")
            return False

        self._validate_value_data(data_value)

        state.timestamp = timestamp

        msg = message_data.MessageData(
            message_data.SEND_REPORT,
            data=str(data_value),
            network_id=state.parent.parent.parent.uuid,
            device_id=state.parent.parent.uuid,
            value_id=state.parent.uuid,
            state_id=state.uuid,
            verb=message_data.PUT
        )
        # self.parent.parent.conn.send_data.send_report(msg)
        self.parent.parent.conn.sending_queue.put(msg)

    def _update_delta_period_values(self, data_value):
        if self.period is not None:
            self.__set_timer()
        if self.delta is not None:
            try:
                self.last_update_of_report = float(data_value)
            except ValueError:
                self.last_update_of_report = float("NAN")

    def check_delta_and_period(self, data_value):
        """
        Check if delta and period allows data to be sent.

        Check if value has delta or period, if it has then if it passes
        checks then True is returned, otherwise False is returned.

        Args:
            data_value: the new value.

        Returns:
            True/False indicating the result of operation.

        """
        if self.delta is not None:
            try:
                if isNaN(data_value):
                    raise ValueError("Value is NAN!")
                data_value = float(data_value)
            except ValueError:
                if not isNaN(self.last_update_of_report):
                    return True
                return self.check_period(False)

            if (
                self.last_update_of_report is None
                or not (abs(data_value - self.last_update_of_report) < self.delta)
            ):
                return True

        return self.check_period(False)

    def check_period(self, return_value):
        """
        Check if period allows data to be sent.

        Check if value has period, if it has then if it passes
        checks then True is returned, otherwise False is returned.

        Args:
            return_value: default return value.

        Returns:
            True/False indicating the result of operation.

        """
        if self.period is not None:
            return self.timer_elapsed
        return return_value

    def get_data(self):
        """
        Get value from report state.

        Check if value has a report state if it has return its data else return
        None.

        Returns:
            Value of report state.

        """
        state = self.get_report_state()
        if state is None:
            return None

        return state.data

    def handle_refresh(self):
        """
        Handles the refresh request.

        Calls __call_callback method with input of 'refresh'

        Returns:
            results of __call_callback

        """
        self.__call_callback('refresh')

    def handle_delete(self):
        """
        Handle delete.

        Calls the __call_callback method with initial input of "remove".

        Returns:
            result of __call_callback method.

        """
        self.__call_callback('remove')

    def delete(self):
        """
        Delete this object.

        Sends delete request for this object and removes its reference
        from parent.

        """
        message = message_data.MessageData(
            message_data.SEND_DELETE,
            network_id=self.parent.parent.uuid,
            device_id=self.parent.uuid,
            value_id=self.uuid
        )
        self.parent.parent.conn.sending_queue.put(message)
        self.parent.values.remove(self)
        self.wapp_log.info("Value removed")

    def __call_callback(self, event):
        if self.callback is not None:
            self.callback(self, event)

    def handle_control(self, data_value):
        """
        Handles the control request.

        Sets data value of control_state object, with value provided and calls
        __call_callback method with input of 'set'.

        Args:
            data_value: the new value.

        Returns:
            results of __call_callback

        """
        # TODO(MBK): Check if the value are within range, and with right step.
        self.control_state.data = data_value

        return self.__call_callback('set')

    def __is_number_type(self):
        """
        Validate data type.

        Checks whether the type of the value is number.

        Returns:
        True if the type is number otherwise false.
        boolean

        """
        return self.data_type == "number"

    def __is_string_type(self):
        """
        Validate data type.

        Checks whether the type of the value is string.

        Returns:
        True if the type is string otherwise false.
        boolean

        """
        return self.data_type == "string"

    def __is_blob_type(self):
        """
        Validate data type.

        Checks whether the type of the value is blob.

        Returns:
        True if the type is blob otherwise false.
        boolean

        """
        return self.data_type == "blob"



"""
Status module for the wappsto module.

Status module to keep track of what the module status is.

Attributes:
    STARTING: Starting phase status.
    INSTANTIATING: Instantiating phase status.
    CONNECTING: Connecting phase status.
    RECONNECTING: Reconnecting phase status.
    CONNECTED: Connected phase status.
    INITIALIZING: Initializing phase status.
    STARTING_THREADS: Thread starting phase status.
    RUNNING: Running phase status.
    DISCONNECTING: Disconnecting phase status.

"""
import logging
from .errors import wappsto_errors


STARTING = "Starting"
CONNECTING = "Connecting"
CONNECTED = "Connected"
INITIALIZING = "Initializing"
STARTING_THREADS = "Starting Threads"
RUNNING = "Running"
RECONNECTING = "Reconnecting"
DISCONNECTING = "Disconnecting"


class Status:
    """
    Status tracking class.

    Tracks the status of the wappsto module through global variables that are
    changed at runtime.
    """

    def __init__(self):
        """
        Initialize the Status class.

        Initializes the Status class, which handles changing the program's
        status flag.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.callback = None
        self.current_status = None

    def set_callback(self, callback):
        """
        Set the callback.

        Sets the callback attribute.

        Args:
            callback: Callback reference.

        Raises:
            CallbackNotCallableException: Custom exception to signify invalid
            callback.

        """
        if not callable(callback):
            msg = "Callback method should be a method"
            self.wapp_log.error("Error setting callback: {}".format(msg))
            raise wappsto_errors.CallbackNotCallableException
        self.callback = callback
        self.wapp_log.debug("Callback {} has been set.".format(callback))
        return True

    def is_starting(self):
        """
        Checks if the status is starting.

        Returns:
            True, if state is "STARTING",
            False, Otherwise.
        """
        return self.current_status == STARTING

    def is_connecting(self):
        """
        Checks if the status is connecting.

        Returns:
            True, if state is "CONNECTING",
            False, Otherwise.
        """
        return self.current_status == CONNECTING

    def is_connected(self):
        """
        Checks if the status is connected.

        Returns:
            True, if state is "CONNECTED",
            False, Otherwise.
        """
        return self.current_status == CONNECTED

    def is_initializing(self):
        """
        Checks if the status is initializing.

        Returns:
            True, if state is "INITIALIZING",
            False, Otherwise.
        """
        return self.current_status == INITIALIZING

    def is_starting_threads(self):
        """
        Checks if the status is starting_threads.

        Returns:
            True, if state is "STARTING_THREADS",
            False, Otherwise.
        """
        return self.current_status == STARTING_THREADS

    def is_running(self):
        """
        Checks if the status is running.

        Returns:
            True, if state is "RUNNING",
            False, Otherwise.
        """
        return self.current_status == RUNNING

    def is_reconnecting(self):
        """
        Checks if the status is reconnecting.

        Returns:
            True, if state is "RECONNECTING",
            False, Otherwise.
        """
        return self.current_status == RECONNECTING

    def is_disconnecting(self):
        """
        Checks if the status is disconnecting.

        Returns:
            True, if state is "DISCONNECTING",
            False, Otherwise.
        """
        return self.current_status == DISCONNECTING

    def get_status(self):
        """
        Retrieve current status.

        Retrieves the class' current_status attribute.

        Returns:
            Current status attribute.

        """
        return self.current_status

    def set_status(self, status):
        """
        Set the current status.

        Sets the current status from the global attribute states.

        Args:
            status: Status global attribute.

        """
        self.current_status = status
        if self.callback is not None:
            self.callback(self)



"""
Message log module.

Logs data being sent in case of in connection and sends it when connection is established.

Attributes:
    REMOVE_OLD: Removes the old data.
    REMOVE_RECENT: Doesnt add the most recent data.

    HOUR_PERIOD: Indicates that log should be compacted once an hour.
    DAY_PERIOD: Indicates that log should be compacted once a day.
    MONTH_PERIOD: Indicates that log should be compacted once a month.

"""
import os
import re
import sys
import json
import time
import zipfile
import logging
import datetime
from json.decoder import JSONDecodeError


REMOVE_OLD = 1
REMOVE_RECENT = 2

HOUR_PERIOD = 3
DAY_PERIOD = 2
MONTH_PERIOD = 1

TIME_BETWEEN_LOG_SEND = 0.1  # time to wait before sending next logged message (seconds)


class OfflineEventStorage:
    """
    Offline event storage.

    Saves data not being sent due to not having connection.
    """

    def __init__(self, log_offline, log_location, log_data_limit, limit_action, compression_period):
        """
        Initialize OfflineEventStorage class.

        Sets up message logging enviroment.

        Args:
            log_offline: boolean indicating if data should be logged
            log_location: location of the logs
            log_data_limit: limit of data to be saved in log [in Megabytes]
            limit_action: action to take when limit is reached
            compression_period: period for compressing data [day, hour]

        Raises:
            ServerConnectionException: "Unable to connect to the server.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        self.log_offline = log_offline
        self.log_data_limit = log_data_limit
        self.limit_action = limit_action
        self.compression_period = compression_period

        self.set_location(log_location)

    def set_location(self, log_location):
        """
        Set log location.

        Sets log location and creates log folder if necassary.

        Args:
            log_location: location of the logs.

        """
        self.log_location = log_location
        os.makedirs(self.log_location, exist_ok=True)

    def get_file_path(self, file_name):
        """
        Gets path to the file.

        Concatenates location of the logs and file name.

        Args:
            file_name: name of the file.

        Returns:
            path to the file.

        """
        return os.path.join(self.log_location, file_name)

    def get_log_name(self):
        """
        Gets name of the newest log.

        Creates and returns the name of the log.

        Returns:
            name of the latest log.

        """
        now = datetime.datetime.now()
        if self.compression_period >= MONTH_PERIOD:
            file_name = str(now.year) + "-" + str(now.month)
        if self.compression_period >= DAY_PERIOD:
            file_name += "-" + str(now.day)
        if self.compression_period >= HOUR_PERIOD:
            file_name += "-" + str(now.hour)
        file_name += ".txt"
        return file_name

    def get_logs(self):
        """
        Gets log files in the location.

        Gets all files from directory and return ones that follow log file format.

        Returns:
            list of log file names.

        """
        file_list = os.listdir(self.log_location)
        pattern = "[0-9][0-9][0-9][0-9]-([0-9]|1[0-2])"
        return [file_name for id, file_name in enumerate(file_list) if re.search(pattern, file_name)]

    def compact_logs(self):
        """
        Compacts all logs to save space.

        Uses all logs received from "get_logs" method and compacts the ones that
        are of type text (after compacting the text file is deleted).

        """
        all_logs = self.get_logs()
        text_logs = [file_name for id, file_name in enumerate(all_logs) if re.search(".txt$", file_name)]
        for file_name in text_logs:
            file_path = self.get_file_path(file_name)
            with zipfile.ZipFile(file_path.replace(".txt", ".zip"), "w") as zip_file:
                zip_file.write(file_path, file_name)
            os.remove(file_path)

    def get_oldest_log_name(self):
        """
        Gets the oldest log's name.

        Uses all logs received from "get_logs" method and sorts them, taking the first (oldest).

        Returns:
            name of the file.

        """
        all_logs = self.get_logs()
        all_logs.sort()
        file_name = all_logs[0]
        return file_name

    def get_text_log(self, file_name):
        """
        Gets name of the file.

        Checks if the file is compacted, if it is then it is unzipped and new name is returned,
        otherwise same file is returned.

        Args:
            file_name: name of the file.

        Returns:
            name of the file.

        """
        if re.search(".zip$", file_name):
            file_path = self.get_file_path(file_name)
            with zipfile.ZipFile(file_path, "r") as zip_file:
                zip_file.extractall(self.log_location)
            os.remove(file_path)
            file_name = file_name.replace(".zip", ".txt")
        return file_name

    def remove_data(self, file_name):
        """
        Removes data.

        Removes lines from the file, but if the number of lines to remove
        exceeds lines in file or file is not text format the file is deleted.

        Args:
            file_name: name of the file.

        """
        file_path = self.get_file_path(file_name)
        if not re.search(".txt$", file_name):
            os.remove(file_path)
            self.wapp_log.debug("Removed old data")
            return

        with open(file_path, "r") as file:
            lines = file.readlines()
        if len(lines) > 1:
            with open(file_path, "w") as file:
                file.writelines(lines[1:])
        else:
            os.remove(file_path)
        self.wapp_log.debug("Removed old data")

    def add_message(self, data):
        """
        Add message to log.

        Adds message to log if logging is enabled otherwise writes error.

        Args:
            data: JSON message data.

        """
        if not self.log_offline:
            self.wapp_log.error("Sending while not connected")
            return

        try:
            string_data = json.dumps(data)
            if (self.log_data_limit * 1000000) >= self.get_size(string_data):
                file_name = self.get_log_name()
                file_path = self.get_file_path(file_name)
                if not os.path.isfile(file_path):
                    # compact data if log for this period doesnt exist
                    self.compact_logs()
                with open(file_path, "a") as file:
                    file.write(string_data + " \n")
                self.wapp_log.debug("Raw log Json: {}".format(string_data))
            else:
                self.wapp_log.debug("Log limit exeeded.")
                if self.limit_action == REMOVE_OLD:
                    file_name = self.get_oldest_log_name()
                    self.remove_data(file_name)
                    self.add_message(data)
                elif self.limit_action == REMOVE_RECENT:
                    self.wapp_log.debug("Not adding data")
        except FileNotFoundError:
            msg = "No log file could be created in: {}".format(self.log_location)
            self.wapp_log.error(msg)

    def get_size(self, data):
        """
        Gets size of log folder.

        Method loops through all files and gets their total size in the folder,
        later also adds size of the data you are trying to save.

        Args:
            data: JSON message data.

        Returns:
            Total size of the folder after changes.

        """
        total_size = 0
        for dirpath, dirnames, file_names in os.walk(self.log_location):
            for file_name in file_names:
                file_path = os.path.join(dirpath, file_name)
                # skip if it is link
                if not os.path.islink(file_path):
                    total_size += os.path.getsize(file_path)

        total_size += sys.getsizeof(data)
        return total_size

    def send_log(self, conn):
        """
        Sends log data.

        If logging is enabled reads all saved messages from log and sends them, later emptying log.

        Args:
            conn: reference to ClientSocket object.

        """
        if self.log_offline:
            try:
                log_list = self.get_logs()
                self.wapp_log.debug("Found log files: " + str(log_list))

                for file_name in log_list:
                    file_name = self.get_text_log(file_name)
                    file_path = self.get_file_path(file_name)
                    with open(file_path, "r") as file:
                        lines = file.readlines()
                    for line in lines:
                        try:
                            data = json.loads(line)
                            for data_element in data:
                                if conn.connected:
                                    time.sleep(TIME_BETWEEN_LOG_SEND)
                                    conn.send_data.create_bulk(data_element)
                                else:
                                    raise ConnectionError
                        except JSONDecodeError:
                            error = "Json decoding error while reading : {}".format(line)
                            self.wapp_log.error(error)
                    self.wapp_log.debug("Data sent from file: " + file_path)
                    os.remove(file_path)
            except FileNotFoundError:
                error = "Log directory could not be found: {}".format(self.log_location)
                self.wapp_log.error(error)
            except ConnectionError:
                # todo maybe should remove sent messages from file being read (so it wouldnt be sent twice)
                self.wapp_log.debug("No connection to the server: Logs are no longer being sent")



"""
Rapid Prototyping Python module for wappsto.com.

Stores the Wappsto class functionality.
"""

import inspect
import logging
import os
import signal
import warnings

from threading import Event

from .connection import communication
from .errors import wappsto_errors
from .connection import event_storage
from . import status
from .data_operation import data_manager


RETRY_LIMIT = 5
warnings.warn("Project is deprecated. Please use https://github.com/Wappsto/python-wappsto-iot instead.", DeprecationWarning)

class Object_instantiation:
    """For backward compability. (Deprecated)."""
    def __init__(self):
        """Initialize Object_instantiation."""
        self.status = status


object_instantiation = Object_instantiation()


class Wappsto:
    """
    The main package classs.

    Establishes a connection to the wappsto server, initializes the classes,
    starts a sending/receiving thread.
    """
    __version__ = "1.2.11"

    def __init__(
        self,
        json_file_name,
        abs_config_path=None,
        load_from_state_file=False,
        log_offline=False,
        log_location="logs",
        log_data_limit=10,
        limit_action=event_storage.REMOVE_OLD,
        compression_period=event_storage.DAY_PERIOD
    ):
        """
        Initialize wappsto class.

        Creates a wappsto object which methods are exposed for users to use.
        These are: starting and stopping connection with a server and
        retrieving devices. These methods are enough for users to make
        operations on their network.

        Args:
            json_file_name: name of a json file containing all information
                about a network
            abs_config_path: The absolute path to all the config files.
                If non is given, it will be set to where it was imported from.
                (Default: None)
            load_from_state_file: Defines if the data should be loaded from
                saved files (default: {False})
            log_offline: boolean indicating if data should be logged
                (default: {False})
            log_location: location of the logs
                (default: {"logs"})
            log_data_limit: limit of data to be saved in log [in Megabytes]
                (default: {10})
            limit_action: action to take when limit is reached
                (default: {REMOVE_OLD})
            compression_period: period for compressing data [day, hour]
                (default: {DAY_PERIOD})

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())

        # TODO(Dimitar): Comment on this later.
        stack = inspect.stack()[1][1]
        if abs_config_path:
            self.path_to_calling_file = abs_config_path
        else:
            self.path_to_calling_file = os.path.dirname(os.path.abspath(stack))

        self.connecting = True
        self.event_storage = event_storage.OfflineEventStorage(
            log_offline,
            log_location,
            log_data_limit,
            limit_action,
            compression_period
        )
        self.socket = None
        self.receive_thread = None
        self.send_thread = None
        self.status = status.Status()
        try:
            self.data_manager = data_manager.DataManager(
                json_file_name=json_file_name,
                load_from_state_file=load_from_state_file,
                path_to_calling_file=self.path_to_calling_file
            )
        except FileNotFoundError as fnfe:
            self.wapp_log.error("Failed to open file: {}".format(fnfe))
            self.stop(False)
            raise fnfe

    def get_status(self):
        """
        Wappsto status.

        Retrieves the status class instance which is set to different statuses
        throughout the program execution.

        Returns:
            A reference to the status object instance.

        """
        return self.status

    def get_network(self):
        """
        Wappsto network.

        Retrieves the network class instance.

        Returns:
            A reference to the network object instance.

        """
        return self.data_manager.network

    def get_devices(self):
        """
        Wappsto devices.

        Retrieves the devices list containing instances of devices

        Returns:
            A list of devices.

        """
        return self.data_manager.network.devices

    def get_by_id(self, id):
        """
        Wappsto get by id.

        Retrieves the instance of a class if its id matches the provided one

        Args:
            id: unique identifier used for searching

        Returns:
            A reference to the network/device/value/state object instance.

        """
        return self.data_manager.get_by_id(id)

    def set_status_callback(self, callback):
        """
        Sets callback for status.

        Sets the provided callback to the instance of status.

        Args:
            callback: reference to callback

        """
        self.status.set_callback(callback)

    def set_network_callback(self, callback):
        """
        Sets callback for network.

        Sets the provided callback to the instance of network.

        Args:
            callback: reference to callback

        """
        self.data_manager.network.set_callback(callback)

    def set_value_callback(self, callback):
        """
        Sets callback for values.

        Sets the provided callback to all instances of value.

        Args:
            callback: reference to callback

        """
        for device in self.data_manager.network.devices:
            for value in device.values:
                value.set_callback(callback)

    def get_device(self, name):
        """
        Device reference.

        Finds the device with a specific name attribute based on a given
        string

        Args:
            name: String containing the name attribute to search for.

        Returns:
            A reference to the device object instance.

        """
        for device in self.data_manager.network.devices:
            if name == device.name:
                return device
        else:
            msg = "Device {} not found in {}".format(name, self.data_manager)
            self.wapp_log.warning(msg, exc_info=True)
            self.stop(False)
            raise wappsto_errors.DeviceNotFoundException(msg)

    def start(self, address="wappsto.com", port=11006, automatic_trace=False, blocking=False):
        """
        Start the server connection.

        Creates a socket object instance and passes the address and port to it.

        Args:
            address: Address to connect the service to.
                     (default: "wappsto.com")
            port: Port to connect the address to.
                  (default: 11006)
            automatic_trace: indicates if all messages automaticaly send trace.
            blocking: Wheather or not this call should be a block call.
                      (default: False)
                      If sat to True, it will listen for a SIGTERM or SIGINT,
                      and terminate if those was received.
                      If this option are set, it is not needed to call stop.

        """
        self.status.set_status(status.STARTING)

        self.socket = communication.ClientSocket(
            data_manager=self.data_manager,
            address=address,
            port=port,
            path_to_calling_file=self.path_to_calling_file,
            wappsto_status=self.status,
            automatic_trace=automatic_trace,
            event_storage=self.event_storage
        )

        self.status.set_status(status.CONNECTING)
        try:
            if not self.socket.connect():
                self.socket.reconnect(RETRY_LIMIT, send_reconnect=False)
        except wappsto_errors.ServerConnectionException as ce:
            self.stop(False)
            raise ce

        self.status.set_status(status.INITIALIZING)
        # Initializes the network, and all the subsequent devices, values and
        # states.
        # TODO(Dimitar): Change from generic Exception
        try:
            self.socket.initialize_all()
        except Exception as e:
            self.wapp_log.error("Error initializing: {}".format(e))
            self.stop(False)
            raise e

        self.status.set_status(status.STARTING_THREADS)
        # Starts the sending & receiving threads.
        try:
            self.receive_thread = self.socket.receiving_thread.start()
            self.send_thread = self.socket.sending_thread.start()
        except Exception as e:
            msg = "Error starting threads: {}".format(e)
            self.wapp_log.error(msg, exc_info=True)
            self.stop(False)
            raise e

        self.status.set_status(status.RUNNING)

        if blocking:
            self.keep_running()

    def keep_running(self):
        """
        Keeps wappsto running.

        Waiting for a SIGTERM or SIGINT, or self.terminated to be set,
        before it will exits.

        """
        self.terminated = Event()

        signal.signal(signal.SIGINT, lambda *args: self.terminated.set())
        signal.signal(signal.SIGTERM, lambda *args: self.terminated.set())
        self.wapp_log.info("Waiting terminate request.")

        self.terminated.wait()
        self.wapp_log.info("Terminate request received.")
        self.stop()

    def stop(self, save=True):
        """
        Stop the Wappsto service.

        Closes the connection socket to the server and, if the boolean is set
        to True it saves an instance of the current state.

        Args:
            save: Flag to determine whether runtime instances should be saved.
                (default: {True})

        """
        self.connecting = False
        self.status.set_status(status.DISCONNECTING)
        # Closes the socket connection, if one is established.
        if self.socket:
            self.socket.close()
        if save:
            self.data_manager.save_instance()
        self.wapp_log.info("Exiting...")



"""
The server communication module.

Stores the server communication functionality for Wappsto as well as the
sending and receiving threads.
"""

import os
import socket
import threading
import time
import queue
import ssl
import logging
from . import message_data
from . import receive_data
from . import send_data
from .. import status
from ..errors import wappsto_errors

from .import seluxit_rpc

PACKET_TIMEOUT = 10


class ClientSocket:
    """
    The ClientSocket class that manages connection.

    Stores the sending and receiving threads, certificates as well as connection
    information.
    """

    def __init__(self, data_manager, address, port, path_to_calling_file,
                 wappsto_status, automatic_trace, event_storage):
        """
        Create a client socket.

        Creates a socket instance for the given address and port. Handles
        transfer of data from the instance attributes and methods to the
        specified server. Connection to the server is based on the specified
        address and port.

        Args:
            data_manager: data_manager of DataManager.
            address: Server address.
            port: Server port.
            path_to_calling_file: path to OS directory of calling file.
            wappsto_status: status object.
            handler: instance of handlers.
            automatic_trace: indicates if all messages automaticaly send trace.
            event_storage: instance of event log.

        """
        self.wapp_log = logging.getLogger(__name__)
        self.wapp_log.addHandler(logging.NullHandler())
        self.data_manager = data_manager
        self.path_to_calling_file = path_to_calling_file
        self.ssl_server_cert = os.path.join(path_to_calling_file,
                                            "certificates/ca.crt")
        self.ssl_client_cert = os.path.join(path_to_calling_file,
                                            "certificates/client.crt")
        self.ssl_key = os.path.join(path_to_calling_file,
                                    "certificates/client.key")
        self.address = address
        self.port = port
        self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self.ssl_context.verify_mode = ssl.CERT_REQUIRED
        self.ssl_context.load_cert_chain(self.ssl_client_cert, self.ssl_key)
        self.ssl_context.load_verify_locations(self.ssl_server_cert)
        self.wappsto_status = wappsto_status

        self.receive_data = receive_data.ReceiveData(self)
        self.receiving_thread = threading.Thread(target=self.receive_data.receive_thread)
        self.receiving_thread.setDaemon(True)

        self.send_data = send_data.SendData(self, automatic_trace)
        self.sending_thread = threading.Thread(target=self.send_data.send_thread)
        self.sending_thread.setDaemon(True)

        self.connected = False
        self.reconnect_inprogres = threading.Lock()
        self.sending_queue = queue.Queue(maxsize=0)
        self.event_storage = event_storage
        self.packet_awaiting_confirm = {}
        self.packet_timeout_list = {}
        self.lock_await = threading.Lock()
        self.set_sockets()

        self.data_manager.network.conn = self

    def set_sockets(self):
        """
        Create socket to communicate with server.

        Creates a socket instance and sets the options for communication.
        Passes the socket to the ssl_wrap method
        """
        self.my_raw_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.my_raw_socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_KEEPALIVE,
            1
        )
        if (hasattr(socket, "TCP_KEEPIDLE")
                and hasattr(socket, "TCP_KEEPINTVL")
                and hasattr(socket, "TCP_KEEPCNT")):
            # After 5 idle minutes, start sending keepalives every 1 minutes.
            # Drop connection after 2 failed keepalives
            self.wapp_log.debug("Setting TCP_KEEPIDLE, TCP_KEEPINTVL & TCP_KEEPCNT.")
            self.my_raw_socket.setsockopt(
                socket.SOL_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPIDLE,
                5 * 60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPINTVL,
                60
            )
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_KEEPCNT,
                2
            )

        if hasattr(socket, "TCP_USER_TIMEOUT"):
            self.wapp_log.debug("Setting TCP_USER_TIMEOUT to 30_000ms.")
            self.my_raw_socket.setsockopt(
                socket.IPPROTO_TCP,
                socket.TCP_USER_TIMEOUT,
                30_000
            )
        self.my_socket = self.ssl_wrap()

    def ssl_wrap(self):
        """
        Wrap socket.

        Wraps the socket using the SSL protocol as configured in the SSL
        context, with hostname verification enabled.

        Returns:
        An SSL wrapped socket.

        """
        return self.ssl_context.wrap_socket(
            self.my_raw_socket,
            server_hostname=self.address
        )

    def connect(self):
        """
        Connect to the server.

        Attempts a connection to the server on the provided addres and port.

        Returns:
            A connection flag to denote if the connection was successful or
            not.

        """
        self.connected = False
        try:
            self.wapp_log.info("Trying to Connect.")
            self.my_socket.settimeout(10)
            self.my_socket.connect((self.address, self.port))
            self.connected = True
            self.my_socket.settimeout(None)
            self.wappsto_status.set_status(status.CONNECTED)
            self.wapp_log.info(f"Connected on interface: {self.my_socket.getsockname()[0]}")
            self.send_logged_data()
            return True

        except Exception as e:
            self.wapp_log.error("Failed to connect: {}".format(e))
            return False

    def send_logged_data(self):
        """
        Sends logged data.

        Makes a thread that sends all of the logged data.
        """
        processThread = threading.Thread(target=self.event_storage.send_log, args=(self,))
        processThread.start()

    def initialize_all(self):
        """
        Call initialize_all method in initialize_code module.

        Initializes the object instances on the sending/receiving queue.
        """
        # for device in self.data_manager.network.devices:
        #     for value in device.values:
        #         state = value.get_control_state()
        #         if state is not None:
        #             msg = message_data.MessageData(
        #                 message_data.SEND_CONTROL,
        #                 network_id=state.parent.parent.parent.uuid,
        #                 device_id=state.parent.parent.uuid,
        #                 value_id=state.parent.uuid,
        #                 state_id=state.uuid,
        #                 verb=message_data.GET
        #             )
        #             self.sending_queue.put(msg)

        trace_id = self.send_data.create_trace(self.data_manager.network.uuid)

        message = seluxit_rpc.get_rpc_whole_json(self.data_manager.get_encoded_network(), trace_id)
        self.send_data.create_bulk(message)  # TODO(MBK): This is not conformed.

        msg = "The whole network {} added to Sending queue {}.".format(
            self.data_manager.network.name,
            self.sending_queue
        )
        self.wapp_log.debug(msg)

        self.confirm_initialize_all()

    def _resend(self, _id):
        with self.lock_await:
            data = self.packet_awaiting_confirm[_id]
        self.wapp_log.info("Resending: {}".format(_id))
        self.remove_id_from_confirm_list(_id)
        self.send_data.create_bulk(data)  # NOTE: Are forced to this.

    def add_id_to_confirm_list(self, data):
        """
        Add the message ID to the confirm list.

        Adds the ID of the decoded JSON message to the list of confirmed
        packets. Uses locks to ensure atomicity.

        Args:
            data: JSON communication message data.

        """
        _id = data.get('id')
        timer = threading.Timer(PACKET_TIMEOUT, lambda: self._resend(_id))
        timer.start()
        self.lock_await.acquire()
        self.packet_timeout_list[_id] = timer
        self.packet_awaiting_confirm[_id] = data
        self.lock_await.release()

    def remove_id_from_confirm_list(self, _id):
        """
        Remove the ID from the confirm list.

        Removes the ID of the decoded JSON message from the list of confirmed
        packets. Uses locks to ensure atomicity.

        Args:
            _id: ID to remove from the confirm list.

        """
        self.lock_await.acquire()
        if _id in self.packet_timeout_list:
            self.packet_timeout_list[_id].cancel()
            del self.packet_timeout_list[_id]
        if _id in self.packet_awaiting_confirm:
            del self.packet_awaiting_confirm[_id]
        self.lock_await.release()
        self.poke_send_thread()

    def poke_send_thread(self):
        """Poke the Send thread, to let it know some thing have changed."""
        poke_msg = message_data.MessageData(
            message_data.POKE
        )
        self.sending_queue.put(poke_msg)

    def request_reconnect(self):
        """Reconnect if it is required (Non Blocking)."""
        self.wapp_log.info("Reconnect Requested.")
        if not self.connected and not self.reconnect_inprogres.locked():
            threading.Thread(target=self.reconnect).start()

    def reconnect(self, retry_limit=None, send_reconnect=True):
        """
        Attempt to reconnect.

        Reconnection attempts in the instance of a connection being interrupted.
        """
        with self.reconnect_inprogres:
            if not self.connected:
                self.wappsto_status.set_status(status.RECONNECTING)
                self.connected = False
                attempt = 0
                while not self.connected and (retry_limit is None
                                              or retry_limit > attempt):
                    attempt += 1
                    self.wapp_log.info("Trying to reconnect in 5 seconds")
                    time.sleep(5)
                    self.close()
                    self.set_sockets()
                    self.connect()

                if self.connected is True:
                    self.wapp_log.info("Reconnected with " + str(attempt) + " attempts")
                    if send_reconnect:
                        reconnect = message_data.MessageData(message_data.SEND_RECONNECT)
                        self.sending_queue.put(reconnect)
                else:
                    msg = "Unable to connect to the server[IP: {}, Port: {}]"
                    msg = msg.format(self.address, self.port)
                    raise wappsto_errors.ServerConnectionException(msg)

    def get_object_without_none_values(self, encoded_object):
        """
        Remove objects with None values.

        Gets objects and removes any keys where value is None.

        Args:
            encoded_object: dictionary object.

        """
        # UNSURE(MBK): I feel that there is a more pythonic way of doing this.
        # ERROR(MBK): Is this the source for slowing down the Porcupine, after a month.
        for key, val in list(encoded_object.items()):
            if val is None:
                del encoded_object[key]
            elif isinstance(val, dict):
                self.get_object_without_none_values(val)
                if len(val) == 0:
                    del encoded_object[key]
            elif isinstance(val, list):
                for val_element in val:
                    self.get_object_without_none_values(val_element)
                    if len(val_element) == 0:
                        val.remove(val_element)
                if len(val) == 0:
                    del encoded_object[key]

    def close(self):
        """
        Close the connection.

        Closes the socket object connection.
        """
        self.wapp_log.info("Closing connection...")

        for device in self.data_manager.network.devices:
            for value in device.values:
                if value.timer.is_alive():
                    msg = "Value: {} is no longer periodically sending updates."
                    msg = msg.format(value.uuid)
                    self.wapp_log.debug(msg)
                value.timer.cancel()

        self.connected = False
        if self.my_socket:
            self.my_socket.close()
            self.my_socket = None
        if self.my_raw_socket:
            self.my_raw_socket.close()
            self.my_raw_socket = None

    def confirm_initialize_all(self):
        """
        Confirms that all responses are received.

        Goes through the list saving expected responses and checks if they are
        received.
        """
        while len(self.packet_awaiting_confirm) > 0:
            self.receive_data.receive_message(fail_on_error=True)
